from django.db.models import Field

from .widgets import CloudFileWidget


class CloudFileField(Field):
    def __init__(self, *args, **kwargs) -> None:
        dest = kwargs.pop("dest", None)
        self.widget = CloudFileWidget(dest=dest)
        super().__init__(*args, **kwargs)

    def get_internal_type(self) -> str:
        return "TextField"

    def formfield(self, *args, **kwargs):
        kwargs["widget"] = self.widget
        return super().formfield(*args, **kwargs)



from urllib.parse import urlparse

from django.conf import settings
from django.http import HttpRequest
from django.shortcuts import render
from django.utils.translation import gettext, gettext_lazy as _
from django.views.generic.base import TemplateView


def generate_link(urlstring: str, request: HttpRequest | None) -> str:
    if not urlstring or urlparse(urlstring).scheme:  # 如果为空或有scheme
        return urlstring
    if request:
        app_scheme = request.headers.get("App-Scheme")
        if app_scheme:
            return f"{app_scheme}://{urlstring}"
    return f"/{urlstring}/" if urlstring != "/" else "/"


class WebView(TemplateView):
    name: str | None = None
    title: str | None = None
    endpoint: str | None = ""

    def get(self, request, *args, **kwargs):
        if self.name.endswith("_create"):  # 创建页
            verbose_name = " ".join(gettext(item) for item in self.name.split("_")[::-1])  # NY: 用gettext_lazy会报错
        else:
            verbose_name = _(self.name.replace("y_list", "ies").replace("_list", "s").replace("_detail", ""))
        title = self.title if self.title is not None else verbose_name  # 用 is not None 才能传入空标题
        if self.endpoint is not None:
            version = settings.REST_FRAMEWORK.get("DEFAULT_VERSION", "")
            endpoint = self.endpoint or f"/api/{version}{request.get_full_path()}"
            dataset_endpoint = f' data-endpoint="{endpoint}"' if endpoint else ""
        template_name = self.template_name or f"web/{self.name}.html"
        extras = kwargs
        is_weixin = "MicroMessenger" in request.headers["User-Agent"]
        return render(request, template_name, locals())



import datetime
import time

from django.conf import settings
from django.utils import timezone


def timestamp(date):
    return time.mktime(date.timetuple())


def generate_shard_id(user_id: int) -> int:
    time_offset = int(timestamp(datetime.datetime.now()) - 1510358400)  # 2017-11-11 00:00:00 中国时间
    return user_id | time_offset << 32


def get_object(model, using="default", default=None, **kwargs):
    try:
        return model.objects.using(using).get(**kwargs)
    except model.DoesNotExist:
        return default


def get_user_id(pk: int) -> int:
    """
    需要保证传入的pk都是int
    """
    return pk & 0xFFFFFFFF


def db_master(user_id: int | None = None) -> str:
    """
    需要保证传入的user_id都是int
    """
    if not user_id:
        return "default"
    if user_id < 100001:
        return "default"
    return f"db_{user_id % settings.SHARD_COUNT}"


def db_slave(user_id: int | None = None) -> str:
    suffix = ""
    return f"{db_master(user_id)}{suffix}"  # replica_


def redis_master(user_id: int | None = None) -> int:
    if not user_id:
        return 0
    if user_id < 100001:
        return 0
    return int(user_id / 1000)


def datetime_to_unixtime(date_time) -> int:
    date_string = date_time.strftime("%Y-%m-%d %H:%M:%S")
    date_time = datetime.datetime.strptime(date_string, "%Y-%m-%d %H:%M:%S") + datetime.timedelta(hours=8)
    return int(time.mktime(date_time.timetuple()))


def unixtime_to_datetime(local_time):
    date_string = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(local_time))
    return datetime.datetime.strptime(date_string, "%Y-%m-%d %H:%M:%S")


def unixtime_to_date(local_time):
    return time.strftime("%Y-%m-%d", time.localtime(local_time))


def datetime_timezone_zero():
    return datetime.datetime(timezone.now().year, timezone.now().month, timezone.now().day, 0, 0, 0)


def string_to_unixtime(string) -> int:
    date_time = datetime.datetime.strptime(string, "%Y-%m-%d %H:%M:%S")
    return int(time.mktime(date_time.timetuple()))



import datetime
import hashlib
import hmac
import mimetypes
import uuid
from base64 import b64encode
from urllib.parse import urlparse

from django.conf import settings
from rest_framework.exceptions import APIException, ValidationError
from rest_framework.permissions import AllowAny, IsAuthenticated
from rest_framework.response import Response
from rest_framework.utils import json
from rest_framework.views import APIView

from ..django.db.utils import timestamp
from .serializers import DownloadUrlSerializer, UploadParamsSerializer


def create_filename(filename) -> str:
    ext = filename.split(".")[-1]
    return f"{uuid.uuid4().hex}.{ext}"


def sign(key, msg):
    return hmac.new(key, msg.encode("utf-8"), hashlib.sha256).digest()


def get_signing_key(key, date_stamp, region, service):
    kDate = sign(("AWS4" + key).encode("utf-8"), date_stamp)
    kRegion = sign(kDate, region)
    kService = sign(kRegion, service)
    return sign(kService, "aws4_request")


def get_signature(msg, digestmod):
    if not settings.CLOUD_STORAGE_SECRET:
        msg = "CLOUD_STORAGE_SECRET 不存在"
        raise APIException(msg)
    key = settings.CLOUD_STORAGE_SECRET.encode()
    return b64encode(hmac.new(key, msg.encode("utf-8"), digestmod).digest()).decode()


def get_params(cloud, region, bucket, filename, rename, expiration, content_encoding, cache_control):
    content_type = mimetypes.guess_type(filename)[0] or "application/octet-stream"
    path = f"upload/{create_filename(filename)}" if rename else filename  # 是否重命名
    # 计算policy
    conditions = [
        {"bucket": bucket},
        ["starts-with", "$key", path.split("/")[0]],
        ["starts-with", "$Content-Type", content_type]
    ]
    policy_dict = {
        "expiration": (datetime.datetime.utcnow() + datetime.timedelta(hours=expiration)).strftime(
            "%Y-%m-%dT%H:%M:%S.000Z"
        ),
        "conditions": conditions
    }
    # 时间相关
    t = datetime.datetime.utcnow()
    expire = t + datetime.timedelta(days=1)
    amz_date = t.strftime("%Y%m%dT%H%M%SZ")
    date_stamp = t.strftime("%Y%m%d")
    if cloud == "aws":
        conditions += [
            {"acl": "public-read"},
            {"success_action_status": "204"},
            {"x-amz-meta-uuid": "14365123651274"},
            {"x-amz-server-side-encryption": "AES256"},
            ["starts-with", "$x-amz-meta-tag", ""],
            {"x-amz-credential": f"{settings.CLOUD_STORAGE_ID}/{date_stamp}/{region}/s3/aws4_request"},
            {"x-amz-algorithm": "AWS4-HMAC-SHA256"},
            {"x-amz-date": amz_date}
        ]
        policy_dict["conditions"] = conditions
        # policy_dict["conditions"].append(["content-length-range", 1, 1024 * 1024 * 4])
    if content_encoding == "gzip":
        policy_dict["conditions"].append(["starts-with", "$Content-Encoding", content_encoding])
    string_to_sign = b64encode(json.dumps(policy_dict).encode("utf-8"))
    params = {"key": path, "Content-Type": content_type, "policy": string_to_sign}
    match cloud:
        case "aliyun":  # 阿里云
            signature = b64encode(
                hmac.new(settings.CLOUD_STORAGE_SECRET.encode("utf-8"), string_to_sign, hashlib.sha1).digest()
            )
            params.update({"OSSAccessKeyId": settings.CLOUD_STORAGE_ID, "signature": signature})
        case "aws":  # AWS https://docs.aws.amazon.com/AmazonS3/latest/API/sigv4-HTTPPOSTConstructPolicy.html
            for condition in conditions:
                if isinstance(condition, dict):
                    params.update(condition)
            params.pop("bucket")
            signing_key = get_signing_key(settings.CLOUD_STORAGE_SECRET, date_stamp, region, "s3")
            signature = hmac.new(signing_key, string_to_sign, hashlib.sha256).hexdigest()  # 16进制
            params.update({"x-amz-meta-tag": "", "x-amz-signature": signature})
    # 其他 Content-Encoding, Cache-Control
    if content_encoding == "gzip":
        params["Content-Encoding"] = content_encoding
    if cache_control:
        params["Cache-Control"] = cache_control
    return params


class DownloadUrlView(APIView):
    """
    获取下载地址
    """

    lookup_field = "cloud"
    serializer_class = DownloadUrlSerializer

    def get(self, request, *args, **kwargs):
        url = request.GET.get(
            "url"
        )  # https://documents-domain-com.oss-cn-shanghai.aliyuncs.com/contract/001/Linux_Command3.pdf
        if not url:
            msg = "url不能为空"
            raise ValidationError(msg)
        url_components = urlparse(url)
        bucket = url_components.netloc.replace(f".{settings.CLOUD_STORAGE_DOMAIN_NAME}", "")
        expires = int(timestamp(datetime.datetime.now())) + 600  # 10分钟有效
        string_to_sign = f"GET\n\n\n{expires}\n/{bucket}{url_components.path}"  # 字符串
        signature = get_signature(string_to_sign, hashlib.sha1)
        params = {"url": f"{url}?OSSAccessKeyId={settings.CLOUD_STORAGE_ID}&Expires={expires}&Signature={signature}"}
        return Response(params)


class UploadParamsView(APIView):
    """
    获取上传参数
    filename -- 文件名, 可含有路径
    expiration -- (可选)过期时间, 默认50年
    content_encoding -- (可选)默认不压缩
    cache_control -- (可选)默认没有
    """

    lookup_field = "cloud"
    serializer_class = UploadParamsSerializer  # 加了只是为OpenAPI用, 不影响输出结果

    def post(self, request, *args, **kwargs):
        filename = request.data.get("filename")
        if not filename:
            msg = "filename不能为空"
            raise ValidationError(msg)
        bucket = request.data.get("bucket", settings.BUCKET_MEDIA)
        expiration = int(request.data.get("expiration", 24 * 365 * 50))  # 过期时间
        content_encoding = request.data.get("content_encoding", "")
        cache_control = request.data.get("cache_control")
        rename = False
        match bucket:
            case settings.BUCKET_MEDIA:
                endpoint = settings.MEDIA_URL
                rename = True
            case settings.BUCKET_STATIC:  # 传到static下的不修改大小写
                endpoint = settings.STATIC_URL
                # filename = filename.lower()
            case settings.BUCKET_CLOUD:
                endpoint = settings.CLOUD_URL
            case _:
                endpoint = ""
        params = get_params(
            kwargs.get(self.lookup_field),
            settings.CLOUD_STORAGE_REGION,
            bucket,
            filename,
            rename,
            expiration,
            content_encoding,
            cache_control
        )
        return Response({"endpoint": endpoint, "params": params})



import re
import shutil
from pathlib import Path

from fabric import task
from invoke import Context


@task
def cleanup(c: Context) -> None:
    """清理"""
    for p in Path().glob("**/*"):
        if p.is_dir():
            if p.name == "__pycache__" or not any(p.iterdir()):
                print(p)
                p.rmdir()
        elif p.name == ".DS_Store" or p.suffix == ".pyc":
            print(p)
            p.unlink()


@task
def distribute(c: Context) -> None:
    """发布到PyPI"""
    cleanup(c)
    # https://packaging.python.org/en/latest/tutorials/packaging-projects/
    c.run("uv build")
    c.run("uv publish --token $UV_PUBLISH_TOKEN")
    shutil.rmtree("dist")


@task
def fix(c: Context) -> None:
    """修复代码"""
    c.run("ruff check --fix")


@task(name="format")
def format_code(c: Context) -> None:
    """格式化代码"""
    c.run("ruff format")
    c.run("uvx isort .")
    for file in [f for d in ["src"] for f in (Path().cwd() / d).rglob("*.py")] + [Path(__file__)]:
        content = file.read_text("utf-8")
        new_content = re.sub(r",\s*\n(\s*)([])}])", r"\n\1\2", content)
        if new_content != content:
            file.write_text(new_content, "utf-8", newline="\n")


@task
def upgrade(c: Context) -> None:
    """升级"""
    remote, local = "https://raw.githubusercontent.com/nyssance/Free/main/", Path(__file__).parent
    c.run(f"curl -fsSL {remote}gitignore/Python.gitignore > {local / ".gitignore"}")
    c.run(f"curl -fsSL {remote}ruff.toml > {local / "ruff.toml"}")



from django.conf import settings
from rest_framework.schemas.openapi import AutoSchema
from rest_framework.schemas.utils import is_list_view


class AutoSchema(AutoSchema):
    def get_operation_id(self, path, method):
        method_name = getattr(self.view, "action", method.lower())
        if is_list_view(path, method, self.view):
            action = "list"
        elif method_name not in self.method_mapping:
            action = self._to_camel_case(method_name)
        else:
            action = self.method_mapping[method.lower()]
        name = self.get_operation_id_base(path, method, action)
        return name + action.capitalize()

    def get_tags(self, path, method):
        if self._tags:
            return self._tags
        path = path.removeprefix("/")
        version = settings.REST_FRAMEWORK.get("DEFAULT_VERSION", "")
        return [path.split("api/")[1].split(f"{version}/")[1].split("/")[0]]
